(function () {
    window.__emmxReverseProxy__ = window.__emmxReverseProxy__ || {}

    function resolveUrlProxy(originalUrl) {

        if (window.__emmxReverseProxy__.proxyTableDirty === true) {
            window.__emmxReverseProxy__.proxyTableDirty = false
            window.__emmxReverseProxy__.proxyTable = JSON.parse(atob(window.__emmxReverseProxy__.proxyTableBase64))
        }

        var mappings = window.__emmxReverseProxy__.proxyTable
        try {
            var originalUrlObj = new URL(originalUrl)
        } catch (e) {
            // If parse url faild, it is a relative url or unrecognized url.
            return null
        }
        var finalUrl = getReverseUrl(originalUrlObj, mappings)
        if (finalUrl) {
            return finalUrl
        }

        // special case: http and https are considered matches
        originalUrlObj.protocol = originalUrlObj.protocol === "https:" ? "http:" : "https:"
        return getReverseUrl(originalUrlObj, mappings)
    }

    function getReverseUrl(originalUrlObj, mappings) {
        for (var i = 0; i < mappings.length; i++) {
            var mapping = mappings[i]

            var internalUrlObj = new URL(mapping.internal.url)
            var externalUrlObj = new URL(mapping.external.url)

            var finalHostname = resolveHostname(originalUrlObj.hostname, internalUrlObj.hostname, externalUrlObj.hostname)
            if (finalHostname === null) {
                continue
            }

            var finalPort = resolveProxyPort(originalUrlObj.port, internalUrlObj.port, externalUrlObj.port)
            if (finalPort === null) {
                continue
            }

            var finalProtocol = resolveProtocol(originalUrlObj.protocol, internalUrlObj.protocol)
            if (finalProtocol === null) {
                continue
            }

            var finalPathname = resolvePathname(originalUrlObj.pathname, internalUrlObj.pathname, externalUrlObj.pathname)
            if (finalPathname === null) {
                continue
            }

            var finalUrlObj = new URL(externalUrlObj.href)
            finalUrlObj.hostname = finalHostname
            if (finalUrlObj.port !== '') {
                // If set port with empty string, it will be 0.
                finalUrlObj.port = finalPort
            }
            finalUrlObj.protocol = finalProtocol
            finalUrlObj.pathname = finalPathname
            finalUrlObj.search = originalUrlObj.search
            finalUrlObj.hash = originalUrlObj.hash
            return finalUrlObj.href
        }
        return null
    }

    function resolveHostname(originalHostname, internalHostname, externalHostname) {
        var userSubdomains = originalHostname.split('.')
        var internalSubdomains = internalHostname.split('.')
        if (internalSubdomains.length > userSubdomains.length) {
            return null
        }

        for (var i = internalSubdomains.length; i > 0; i--) {
            if (userSubdomains[userSubdomains.length - 1].toLocaleLowerCase() == internalSubdomains[i - 1].toLocaleLowerCase()) {
                userSubdomains.pop()
                continue
            }

            // In javascript, '*' may be encoded to '%2A'
            if (i == 1 && (internalSubdomains[i - 1] == '*' || internalSubdomains[i - 1] == '%2A')) {
                break
            }
            return null
        }

        var finalSubdomains = userSubdomains
        var externalSubdomains = externalHostname.split('.')
        if (externalSubdomains.length > 0 && (externalSubdomains[0] == '*' || externalSubdomains[0] == '%2A')) {
            externalSubdomains.splice(0, 1)
        }

        finalSubdomains = finalSubdomains.concat(externalSubdomains)

        return finalSubdomains.join('.')
    }

    function resolveProxyPort(originalPort, internalPort, externalPort) {
        //Case: originalPort and internalPort are equal (or both nil). This is considered a match.
        //In this case, externalPort replaces self.port
        if (originalPort == internalPort) {
            return externalPort
        }

        //Case: internalPort is non-nil and not equal to originalPort. This is not considered a match.
        //In javascript, non port is empty string.
        if (internalPort !== "") {
            return null
        }

        //The remaining case: internalPort is nil and originalPort is non-nil. This is considered
        //a match. In this case, originalPort takes precedence over externalPort, i.e.
        //originalPort comes back unchanged
        return originalPort
    }

    function resolvePathname(orinalPath, internalPath, externalPath) {
        var userPathComponents = orinalPath.split('/').slice(1) // ignore first empty
        var internalPathComponents = internalPath.split('/').slice(1)

        // ignore last empty
        if (userPathComponents[userPathComponents.length - 1] === '') {
            userPathComponents.pop()
        }
        if (internalPathComponents[internalPathComponents.length - 1] === '') {
            internalPathComponents.pop()
        }

        if (userPathComponents.length < internalPathComponents.length) {
            return null
        }
        for (let i = 0; i < internalPathComponents.length; i++) {
            if (userPathComponents[0] !== internalPathComponents[i]) {
                return null
            }
            userPathComponents.splice(0, 1)
        }
        var finalPathComponents = externalPath.split('/')
        if (finalPathComponents[finalPathComponents.length - 1] == "") {
            finalPathComponents.pop()
        }

        finalPathComponents = finalPathComponents.concat(userPathComponents)
        return finalPathComponents.join('/')
    }

    function resolveProtocol(originalProtocol, internalProtocol) {
        if (originalProtocol !== internalProtocol) return null;
        return "https:";
    }

    // For XMLHTTPRequest
    var __xhrOpen = __xhrOpen || XMLHttpRequest.prototype.open;
    XMLHttpRequest.prototype.open = function (method, url, async, user, pass) {
        var args = arguments
        if (args.length >= 2) {
            var proxyUrl = resolveUrlProxy(url)
            if (proxyUrl) {
                args[1] = proxyUrl
            }
        }
        __xhrOpen.apply(this, args)
    };

    // For fetch api
    var __windowFetch = __windowFetch || window.fetch
    window.fetch = function (resource, init) {
        var proxyResource = null
        if (resource && resource.url) {
            // resource is Request object
            var proxyUrl = resolveUrlProxy(resource.url)
            if (proxyUrl !== null) {
                proxyResource = resource.clone();
                proxyResource.url = proxyUrl
            } else {
                proxyResource = resource
            }
        } else {
            // resource is string
            var proxyUrl = resolveUrlProxy(resource)
            if (proxyUrl !== null) {
                proxyResource = proxyUrl
            } else {
                proxyResource = resource
            }
        }

        return __windowFetch.call(this, proxyResource, init)
    }

    // ----- For HTML tag url proxy ----- Begin
    var proxyNodeList = ['img', 'script', 'link', 'video', 'audio', 'source', 'iframe', 'input'] //Tag name should be lower case here
    var mutationObserver = new MutationObserver(function (mutationList) {
        for (var mutation of mutationList) {
            if (mutation.type === 'childList') {
                //Document added new nodes
                for (var node of mutation.addedNodes) {
                    replaceNodeUrlIfNeed(node)
                }
            } else if (mutation.type === 'attributes') {
                //Nodes attributes changed
                replaceNodeUrlIfNeed(mutation.target)
            }
        }
    })

    function replaceNodeUrlIfNeed(node) {
        if (!isProxyNode(node.nodeName)) {
            return
        }

        var originalSrc = node.getAttribute('src')
        var originalHref = node.getAttribute('href')

        if (originalSrc !== null && (originalSrc.indexOf("http://") === 0 || originalSrc.indexOf("https://") === 0)) {
            var proxySrc = resolveUrlProxy(originalSrc)
            if (proxySrc !== null && proxySrc !== originalSrc) {
                node.setAttribute('src', proxySrc)
            }
        }


        if (originalHref !== null && (originalHref.indexOf("http://") === 0 || originalHref.indexOf("https://") === 0)) {
            var proxyHref = resolveUrlProxy(originalHref)
            if (proxyHref !== null && proxyHref !== originalHref) {
                node.setAttribute('href', proxyHref)
            }
        }
    }

    function isProxyNode(nodeName) {
        var nodeNameLowerCase = nodeName.toLowerCase()
        for (var proxyNodeName of proxyNodeList) {
            if (nodeNameLowerCase === proxyNodeName) {
                return true
            }
        }
        return false
    }

    window.__emmxReverseProxy__.shouldProxyHTMLTags = function(url) {
        if (window.__emmxReverseProxy__.proxyTableDirty === true) {
            window.__emmxReverseProxy__.proxyTableDirty = false
            window.__emmxReverseProxy__.proxyTable = JSON.parse(atob(window.__emmxReverseProxy__.proxyTableBase64))
        }

        var mappings = window.__emmxReverseProxy__.proxyTable
        var reverseMappings = []
        for(var mapping of mappings) {
            reverseMappings.push({
                external:{ url: mapping.internal.url},
                internal:{ url: mapping.external.url}
            })
        }
        var currentUrl = new URL(url)
        var matchUrl = getReverseUrl(currentUrl, reverseMappings)
        if (matchUrl) {
            return true
        }

        // special case: http and https are considered matches
        currentUrl.protocol = currentUrl.protocol === "https:" ? "http:" : "https:"
        matchUrl = getReverseUrl(currentUrl, reverseMappings)
        if (matchUrl) {
            return true
        }

        return false
    }

    window.addEventListener('DOMContentLoaded', function () {
        if (!window.__emmxReverseProxy__.shouldProxyHTMLTags(window.location.href)) {
            return
        }
        
        for (var nodeName of proxyNodeList) {
            var nodes = window.document.getElementsByTagName(nodeName)
            for (var node of nodes) {
                replaceNodeUrlIfNeed(node)
            }
        }

        mutationObserver.observe(window.document, { childList: true, subtree: true, attributes: true, attributeFilter: ['src', 'href'] })
    })
    // ----- For HTML tag url proxy ----- End
})()
